// $Id: RowGrid.cpp,v 1.23 2006/09/08 01:15:29 darseq Exp $
// copyright by PostWare (S.B.M. Post)
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#pragma resource "*.res"

#include <algorithm>
#include "RowGrid.h"
#ifdef _DEBUG
#define PTRSIZE 8+1
// +1: '\0' character
#include "Exceptions.h"
#endif
#define MOVEDELTA 2
// when dragging cells
#pragma package(smart_init)
//---------------------------------------------------------------------------
// ValidCtrCheck is used to assure that the components created do not have
// any pure virtual functions.
static inline void ValidCtrCheck(TRowGrid *) {
	new TRowGrid(NULL);
}
//---------------------------------------------------------------------------

namespace Rowgrid {
	void __fastcall PACKAGE Register() {
		TComponentClass classes[1] = {__classid(TRowGrid)};
		RegisterComponents("TRowGrid", classes, 0);
	}
}
//---------------------------------------------------------------------------

// todo: RowGrid: add scroll support. i.e when Units*UnitWidth > ClientWidth
__fastcall TRowGrid::TRowGrid(TComponent* Owner) : TCustomControl(Owner) {
	internal_call = dragging = false;
#ifndef DEMO
	Caption = "Copyright 2004-2006 by PostWare (S.B.M. Post), "
		"01234567890123456789012345678901"; // 32 characters
#else
	Caption = "Copyright 2004-2006 by PostWare (S.B.M. Post), Demo version";
#endif

	FDragState = dsNone;

	Units = 36; RowCount = 10; RowHeight = 20;
	RowLineWidth = 10; UnitWidth = 15; CellBorderWidth = 8;
	FrameWidth = 2; CellFrameWidth = 1;
	Width = Units*UnitWidth + 2*FrameWidth;
	Height = (RowCount+1)*RowLineWidth + RowCount*RowHeight + 2*FrameWidth;

	DefaultDrawing = true;
	CheckPosition = true;
	CrossBorders = true;
	VerticalRows = false;
	AllowAltKey = false;
	AllowShiftKey = true;
	MinVisible = 1;
	RowLineColor = clSkyBlue;
	CellColor = clBackground;
	TickColor = clBlack;
	OnInsertRowCell = NULL;
	OnRemoveRowCell = NULL;
	OnClearRowCell = NULL;
	OnDrawRowLine = NULL;
	OnDrawRowCell = NULL;
	OnDragRowCell = NULL;
//	OnMouseOverRowCell = NULL;
	OnDblClickRowCell = NULL;

	// what if multiple TRowGrid instances?
	Screen->Cursors[crArrowLeft] = LoadCursor(HInstance,"ArrowLeft");
	Screen->Cursors[crArrowRight] = LoadCursor(HInstance,"ArrowRight");
	Screen->Cursors[crArrowTop] = LoadCursor(HInstance,"ArrowTop");
	Screen->Cursors[crArrowBottom] = LoadCursor(HInstance,"ArrowBottom");
	Screen->Cursors[crArrowHoriz] = LoadCursor(HInstance,"ArrowHoriz");
	Screen->Cursors[crArrowVert] = LoadCursor(HInstance,"ArrowVert");
}
//---------------------------------------------------------------------------

void __fastcall TRowGrid::ClearRow(int Row) {
	row_t & cells = Cells[Row];
	row_t::iterator cell = cells.begin();
	while(cell != cells.end()) {
		bool can_clear = true;
		if(OnClearRowCell) {OnClearRowCell(this,*cell,can_clear);}
		if(can_clear) {cell = cells.erase(cell);}
		else {cell++;}
	}
}
//---------------------------------------------------------------------------

void __fastcall TRowGrid::insertRowCell(int row,int start,int units,
	int zorder,void * object) {
	internal_call = true; InsertRowCell(row,start,units,zorder,object);
}
//---------------------------------------------------------------------------

void __fastcall TRowGrid::removeRowCell(RowCell & row_cell) {
	internal_call = true; RemoveRowCell(row_cell);
}
//---------------------------------------------------------------------------

void __fastcall TRowGrid::InsertRowCell(int Row,int Start,int Units,int Zorder,
	void * Object) {
	if(Row < 0 || Row >= RowCount) {
		throw Exception("RowGrid::InsertRowCell, row index out of bounds");
	}
	if(Units < 0 || Zorder < 0) {
		throw Exception("RowGrid::InsertRowCell, invalid Units/Zorder value");
	}
	RowCell insert_cell;
	insert_cell.FRow = Row;
	insert_cell.FStart = Start;
	insert_cell.FUnits = Units;
	insert_cell.FZorder = Zorder;
	insert_cell.Object = Object;

	can_insert = true;
	if(internal_call) {
		internal_call = false;
		// clear internal_call as soon as possible:
		// suppose exception in OnInsertRowCell event
#ifdef _DEBUG
		AnsiString message("OnInsertRowCell");
		message += " row:"+IntToStr(insert_cell.Row);
		message += " start:"+IntToStr(insert_cell.Start);
		message += " units:"+IntToStr(insert_cell.Units);
		message += " zorder:"+IntToStr(insert_cell.Zorder);
		char string[PTRSIZE];
		snprintf(string,PTRSIZE,"%x",insert_cell.Object);
		message += " object:"+String(string);
		Exceptions::writeLog(message);
#endif
		if(OnInsertRowCell) {OnInsertRowCell(this,insert_cell,can_insert);}
	}

	// After OnInsertRowCell: The cell may have been
	// supplied with an object or can_insert was cleared
	if(can_insert) {Cells[Row].push_back(insert_cell);}
}
//---------------------------------------------------------------------------

void __fastcall TRowGrid::RemoveRowCell(RowCell & RowCell) {
	if(RowCell.Row < 0 || RowCell.Row >= RowCount) {
		throw Exception("RowGrid::RemoveRowCell, row index out of bounds");
	}
	row_t::iterator cell = Cells[RowCell.Row].begin();
	row_t::iterator end = Cells[RowCell.Row].end();
	for(;cell != end;cell++) {if(*cell == RowCell) {break;}}
	// break: when identical cells, delete the first one encountered

	if(cell == end) {
		throw Exception("RowGrid::RemoveRowCell, cell not present");
	}

	can_remove = true;
	if(internal_call) {
		internal_call = false;
		// clear internal_call as soon as possible:
		// suppose exception in OnRemoveRowCell event
#ifdef _DEBUG
		AnsiString message("OnRemoveRowCell");
		message += " row:"+IntToStr(cell->Row);
		message += " start:"+IntToStr(cell->Start);
		message += " units:"+IntToStr(cell->Units);
		message += " zorder:"+IntToStr(cell->Zorder);
		char string[PTRSIZE];
		snprintf(string,PTRSIZE,"%x",cell->Object);
		message += " object:"+String(string);
		Exceptions::writeLog(message);
#endif
		// User is responsible for releasing cell->object. When
		// internal_call this should be done in OnRemoveCell.
		if(OnRemoveRowCell) {OnRemoveRowCell(this,*cell,can_remove);}
	}

	// After OnRemoveRowCell: The cell object must
	// first be deleted or can_remove was cleared.
	if(can_remove) {Cells[RowCell.Row].erase(cell);}
}
//---------------------------------------------------------------------------

void __fastcall TRowGrid::Paint(void) {
#ifdef _DEBUG
	AnsiString message("Paint");
	Exceptions::writeLog(message);
#endif
	bool restore_dragging = dragging;
	cells_t::iterator row = Cells.begin();
	for(int r = 0;r <= RowCount;r++) {
		if(r != RowCount) {
			sort(row->begin(),row->end());
			row_t::iterator cell = row->begin();
			for(;cell != row->end();cell++) {PaintCell(*cell);}
			row++;
		}
		PaintRow(r);
	}
	dragging = restore_dragging;

	TRect frame(0,0,Width,Height);
	Frame3D(Canvas,frame,clBlack,clBlack,FrameWidth);
	if(DragState != dsNone) {draw_twice = false; drawFocusRect(old_rect);}
}
//---------------------------------------------------------------------------

void __fastcall TRowGrid::PaintCell(RowCell & row_cell) {
	TRect rect; int clientwidth = ClientWidth;
	rect.Left = FrameWidth + row_cell.Start*UnitWidth;
	rect.Right = rect.Left + row_cell.Units*UnitWidth;
	rect.Top = FrameWidth + RowLineWidth +
		row_cell.Row*(RowLineWidth+RowHeight);
	rect.Bottom = rect.Top + RowHeight;

	if(VerticalRows) {clientwidth = ClientHeight;}
	if(rect.Left >= FrameWidth + clientwidth ||
		rect.Right <= FrameWidth) {return;}
	if(VerticalRows) {flipRect(rect);}

	if(DefaultDrawing) {
		Frame3D(Canvas,rect,clBlack,clBlack,CellFrameWidth);
		Canvas->Brush->Color = CellColor;
#ifndef DEMO
		Canvas->FillRect(rect);
#else
		Canvas->TextRect(rect,rect.Left,rect.Top,"DEMO VERSION");
#endif
	}
#ifndef DEMO
	if(OnDrawRowCell) {OnDrawRowCell(this,row_cell,rect);}
#endif
}
//---------------------------------------------------------------------------

void __fastcall TRowGrid::PaintRow(int row) {
	TRect rect;
	rect.Left = FrameWidth;
	rect.Right = rect.Left + Units*UnitWidth;
	rect.Top = FrameWidth + row*(RowLineWidth+RowHeight);
	rect.Bottom = rect.Top + RowLineWidth;
	if(VerticalRows) {flipRect(rect);}

	if(DefaultDrawing) {
		Canvas->Brush->Color = RowLineColor;
		Canvas->FillRect(rect);
		Canvas->Pen->Color = TickColor;
		int middle2 = RowLineWidth/2;
		int middle4 = RowLineWidth/4;
		if(VerticalRows) {
			Canvas->MoveTo(rect.Left + middle2,FrameWidth);
			Canvas->LineTo(rect.Left + middle2,FrameWidth + Units*UnitWidth);
		} else {
			Canvas->MoveTo(FrameWidth,rect.Top + middle2);
			Canvas->LineTo(FrameWidth + Units*UnitWidth,rect.Top + middle2);
		}
		for(int u=1;u < Units;u++) {
			int x = FrameWidth + u*UnitWidth;
			Canvas->Pen->Color = TickColor;
			if(VerticalRows) {
				Canvas->MoveTo(rect.Left+middle4,x);
				Canvas->LineTo(rect.Right-middle4,x);
			} else {
				Canvas->MoveTo(x,rect.Top+middle4);
				Canvas->LineTo(x,rect.Bottom-middle4);
			}
		}
	}
#ifndef DEMO
	if(OnDrawRowLine) {OnDrawRowLine(this,rect,row);}
#endif
}
//---------------------------------------------------------------------------

void __fastcall TRowGrid::flipRect(TRect & rect) {
	int left = rect.Left; int right = rect.Right;
	int top = rect.Top; int bottom = rect.Bottom;
	rect.Left = top; rect.Right = bottom;
	rect.Top = left; rect.Bottom = right;
}
//---------------------------------------------------------------------------

void __fastcall TRowGrid::drawFocusRect(TRect & rect) {
	TRect new_rect = rect; Canvas->Brush->Color = clWhite; // where no dots
	Canvas->Pen->Mode = pmNotXor; Canvas->Pen->Style = psDot;
	if(draw_twice) {Frame3D(Canvas,old_rect,clBlack,clBlack,1);}
	Frame3D(Canvas,new_rect,clBlack,clBlack,1);
	Canvas->Pen->Mode = pmCopy; Canvas->Pen->Style = psSolid;
	old_rect = rect; draw_twice = true;
}
//---------------------------------------------------------------------------

RowCell __fastcall TRowGrid::MouseToCell(int X,int Y,TCellState & State) {
	int x,y; if(VerticalRows) {x = Y; y = X;} else {x = X; y = Y;}

	State = csNone;
	RowCell row_cell;
	row_cell.FRow = 0;
	row_cell.FZorder = 0;
	int old_top = FrameWidth;

	bool search = true;
	bool restore_dragging = dragging;
	cells_t::iterator row = Cells.begin();
	for(int r = 0;search && r <= RowCount;r++) {
		int top = FrameWidth + r*(RowLineWidth+RowHeight);
		if(top + RowLineWidth <= y) { // <=, top & left included
			if(r == RowCount) {break;}
			if(y < top + RowLineWidth + RowHeight) {
				sort(row->begin(),row->end());
				row_t::reverse_iterator cell = row->rbegin();
				for(;search && cell != row->rend();cell++) {
					int left = FrameWidth + cell->Start*UnitWidth;
					int right = left + cell->Units*UnitWidth;
					// <=, top & left included
					if(left <= x && x < right) {
						search = false;
						State = csCenter;
						if(x-left <= CellBorderWidth) {State = csLeft;}
						if(right-x <= CellBorderWidth) {State = csRight;}
						row_cell = *cell;
					}
				}
				if(search) {
					search = false;
//					if(FrameWidth < x && x < FrameWidth + Units*UnitWidth) {
						row_cell.FRow = r; State = csEmpty;
//					}
				}
			}
			row++; // r != RowCount
		} else {
			search = false;
			if(r && old_top + RowLineWidth <= y && y < top) {
//				&& FrameWidth < x && x < FrameWidth + Units*UnitWidth) {
				row_cell.FRow = r-1; State = csEmpty;
			}
		}
		old_top = top;
	}
	dragging = restore_dragging;

#ifdef _DEBUG
	AnsiString message("MouseToCell");
	message += " X:"+IntToStr(X);
	message += " Y:"+IntToStr(Y);
	message += " row:"+IntToStr(row_cell.Row);
	message += " start:"+IntToStr(row_cell.Start);
	message += " units:"+IntToStr(row_cell.Units);
	message += " zorder:"+IntToStr(row_cell.Zorder);
	message += " state:"+IntToStr(State);

	char string[PTRSIZE];
	snprintf(string,PTRSIZE,"%x",row_cell.Object);
	message += " object:"+String(string);
	Exceptions::writeLog(message);
#endif

	return row_cell;
}
//---------------------------------------------------------------------------

bool __fastcall TRowGrid::MouseToShift(int X,int Y,int & old_start,
	int & new_start) {
	// determine the new position (new_start) of the cell being dragged
	// determine the old position (old_start) of the cell being dragged
	// new_start - old_start: howmuch the mouse has shifted in units
	//     csCenter          csCenter          csRight           csLeft
	//       dx--->X        X<---dx              dx--->X        X<---dx
	//  o_____|_   |        |  o_|_____     o_____|_ __|_      _|__o__|____
	//  |--->n__|__|_     n_|__|__     |    |--->n__|____|   n|____|__     |
	//  |____|__|    |    |<---|__|____|    |____|__|____|    |<---|__|____|
	//       |_______|    |_______|              |_______|    |_______|
	// ---------------------------------------------------------------------
	int mouse_pos,check_pos,drag_pos;
	if(VerticalRows) {mouse_pos = check_pos = Y; drag_pos = drag_y;}
	else {mouse_pos = check_pos = X; drag_pos = drag_x;}
	if(DragState != dsInsert) {
		int left = FrameWidth + drag_cell.Start*UnitWidth;
		mouse_pos = mouse_pos - drag_pos + left;
	}
	new_start = RoundTo((mouse_pos - FrameWidth) * 1.0/UnitWidth,0);
	if(DragState == dsInsert) {
		old_start = RoundTo((drag_pos - FrameWidth) * 1.0/UnitWidth,0);
		if(new_start < old_start) {cell_state = csLeft;}
		else {cell_state = csRight;}
	} else {old_start = drag_cell.Start;}

	// the new position (new_start) may not be valid. If CheckPosition is true,
	// we check for overlap with other cells and change new_start if necessary
	bool pos_ok = true;
	if(chk_position) {
		if(DragState == dsInsert || cell_state != csCenter) {
			pos_ok = checkPosition(drag_pos,old_start,new_start);
		} else {pos_ok = checkPosition(check_pos,old_start,new_start);}
	}

	// prevent dragging one cellborder over the other
	if(DragState != dsInsert) {
		if(cell_state == csLeft) {
			if(new_start > old_start+drag_cell.Units-1) {
				new_start = old_start+drag_cell.Units-1;
			}
		} else if(cell_state == csRight) {
			if(new_start < old_start-drag_cell.Units+1) {
				new_start = old_start-drag_cell.Units+1;
			}
		}
	}

#ifdef _DEBUG
	AnsiString message("MouseToShift");
	message += " old_start:"+IntToStr(old_start);
	message += " new_start:"+IntToStr(new_start);
	message += " DragState:"+IntToStr(DragState);
	message += " pos_ok:"+IntToStr(pos_ok);
	Exceptions::writeLog(message);
#endif

	return pos_ok;
}
//---------------------------------------------------------------------------

bool __fastcall TRowGrid::checkPosition(int X,int & old_start,int & new_start) {
	// find markers m1 and m2 around X, X must be a csEmpty position
	int m1 = 0; int m2 = Units;
	row_t::iterator min; bool min_set = false;
	row_t::iterator max; bool max_set = false;
	row_t::iterator cell = Cells[drag_cell.Row].begin();
	row_t::iterator end = Cells[drag_cell.Row].end();
	for(;cell != end;cell++) {
		int min_r; int max_l; // |l min r| <= X < |l max r|
		int left = FrameWidth + cell->Start*UnitWidth;
		int right = left + cell->Units*UnitWidth;
		if(right <= X) {
			if(min_set == false) {min_set = true; min = cell; min_r = right;}
			else if(min_r < right) {min = cell; min_r = right;}
		}
		if(left > X) {
			if(max_set == false) {max_set = true; max = cell; max_l = left;}
			else if(max_l > left) {max = cell; max_l = left;}
		}
	}
	if(min_set) {m1 = min->Start + min->Units;}
	if(max_set) {m2 = max->Start;}

#ifdef _DEBUG
	AnsiString message("checkPosition");
	message += " X:"+IntToStr(X);
	message += " m1:"+IntToStr(m1);
	message += " m2:"+IntToStr(m2);
	Exceptions::writeLog(message);
#endif

	// check if new_start fits between m1 and m2
	int units = 0; if(DragState != dsInsert) {units = drag_cell.Units;}
	int lim = MinVisible; if(lim > units) {lim = units;}
	bool drop_ok = true;
	if(cell_state == csLeft) { // dsInsert/Resize
		if(old_start+units > Units &&
			new_start > Units-lim) {new_start = Units-lim;}
		if(new_start < m1) {new_start = m1;}
	} else if(cell_state == csRight) { // dsInsert/Resize
		if(old_start < 0 &&
			new_start < lim-units) {new_start = lim-units;}
		if(new_start > m2-units) {new_start = m2-units;}
	} else { // dragging and dropping onto another location
		if(new_start < lim-units) {new_start = lim-units;}
		if(new_start < m1) {
			if(m1 != 0) {drop_ok = false;} // cell does not fit
			else if(CrossBorders == false) {new_start = m1;}
		}
		if(new_start > Units-lim) {new_start = Units-lim;}
		if(new_start > m2-units) {
			if(m2 != Units) {drop_ok = false;} // cell does not fit
			else if(CrossBorders == false) {new_start = m2-units;}
		}
	}
	return drop_ok;
}
//---------------------------------------------------------------------------

void __fastcall TRowGrid::MouseMove(TShiftState Shift,int X,int Y) {
	Update(); // Invalidate(), MouseMove, Paint sequence is invalid. Paint first
#ifndef _DEBUG
	TCellState move_state;
	MouseToCell(X,Y,move_state);
//	TCellState move_state = moveMouse(X,Y);
#endif
	if(dragging && drag_hold == false &&
		abs(X-drag_x) < MOVEDELTA &&
		abs(Y-drag_y) < MOVEDELTA) {TControl::MouseMove(Shift,X,Y); return;}
		// when MOVEDELTA=1, this is still possible! (ActiveX only?)

	if((dragging && cell_state != csNone) || DragState != dsNone) {
		// Set DragState in move and not in down: before dblclick, down
		// is called which would lead to an undesired state change. It may
		// still be however that move changes the state before a dblclick
		// in which case the dblclick is ignored by setting dblclick = false
		if(dragging/*&& DragState == dsNone*/) { // dragging -> DragState=dsNone
			dragging = false; dblclick = false; // also see MouseDown
			// Do not try over and over again if state transition fails once.
			// This also ensures that MouseDown was called and drag_cell was
			// properly set. To ensure that drag_cell represents a valid cell,
			// dragging is cleared when the RowGrid Cells property is accessed
			TDragState state = dsNone;
			if(cell_state == csEmpty) {state = dsInsert;}
			else if(drag_button == mbLeft) {state = dsDragDrop;}
			else if(drag_button == mbRight) {state = dsCopyDrop;}
			if(state == dsDragDrop || state == dsCopyDrop) {dragRowCell(state);}
			FDragState = state; // set after dragRowCell <-> event exception
		}

		if(DragState != dsNone) {
			// [When cell remains at its position, the mouse cell sometimes is
			// placed at the wrong position. This is caused by x which is not
			// being properly converted back to units: units -> x -> units+/-1]
			TRect new_rect; int old_start,new_start;
			MouseToShift(X,Y,old_start,new_start);
			int x = FrameWidth + new_start*UnitWidth;
			int height = RowHeight;
			int top = FrameWidth + RowLineWidth +
				drag_cell.Row*(RowLineWidth+height);
			if(DragState == dsInsert) {
				int x_old = FrameWidth + old_start*UnitWidth;
				if(x_old < x) {new_rect = TRect(x_old,top,x,top+height);}
				else {new_rect = TRect(x,top,x_old,top+height);}
			} else { // dsDragDrop/dsCopyDrop
				int left = FrameWidth + drag_cell.Start*UnitWidth;
				int right = left + drag_cell.Units*UnitWidth;
				int width = right - left;
				if(cell_state == csLeft) {
					new_rect=TRect(x,top,right,top+height);
				} else if(cell_state == csRight) {
					new_rect=TRect(left,top,x+width,top+height);
				} else if(VerticalRows) {
					new_rect=TRect(x,X-drag_x+top,x+width,X+height-drag_x+top);
				} else {
					new_rect=TRect(x,Y-drag_y+top,x+width,Y+height-drag_y+top);
				}
			}
			if(VerticalRows) {flipRect(new_rect);}
			drawFocusRect(new_rect);
		}
	} else {
		TCursor cursor = crDefault;
#ifndef _DEBUG
		if(move_state == csLeft) {
			if(VerticalRows) {cursor = (TCursor) crArrowBottom;}
			else {cursor = (TCursor) crArrowRight;}
		} else if(move_state == csRight) {
			if(VerticalRows) {cursor = (TCursor) crArrowTop;}
			else {cursor = (TCursor) crArrowLeft;}
		}
#endif
		Screen->Cursor = cursor;
	}
	TControl::MouseMove(Shift,X,Y);
}
//---------------------------------------------------------------------------

/*
TCellState __fastcall TRowGrid::moveMouse(int X,int Y) {
	TCellState move_state; RowCell move_cell;
	move_cell = MouseToCell(X,Y,move_state);
	if(move_state != csNone && move_state != csEmpty) {
		// todo: && (prev_cell == move_cell) == false
		// prev_cell = move_cell;
		// how should prev_cell be cleared when its matching
		// RowCell is cleared and how to initialize prev_cell
//#ifdef _DEBUG
//		AnsiString message("OnMouseOverRowCell");
//		message += " row:"+IntToStr(move_cell.Row);
//		message += " start:"+IntToStr(move_cell.Start);
//		message += " units:"+IntToStr(move_cell.Units);
//		Exceptions::writeLog(message);
//#endif
		if(OnMouseOverRowCell) {OnMouseOverRowCell(this,move_cell);}
	}
	return move_state;
}
//---------------------------------------------------------------------------
*/

void __fastcall TRowGrid::dragRowCell(Rowgrid::TDragState & new_state) {
#ifdef _DEBUG
	AnsiString message("OnDragRowCell");
	message += " row:"+IntToStr(drag_cell.Row);
	message += " start:"+IntToStr(drag_cell.Start);
	message += " units:"+IntToStr(drag_cell.Units);
	message += " zorder:"+IntToStr(drag_cell.Zorder);

	char string[PTRSIZE];
	snprintf(string,PTRSIZE,"%x",drag_cell.Object);
	message += " object:"+String(string);
	Exceptions::writeLog(message);
#endif
	if(new_state == dsCopyDrop && cell_state != csCenter) {
		// resizing is not allowed when the new state equals dsCopyDrop.
		new_state = dsNone;
	} else {
		if(OnDragRowCell) {
			bool can_drag = true;
			OnDragRowCell(this,drag_cell,cell_state,
				new_state == dsCopyDrop,can_drag);
			if(can_drag == false) {new_state = dsNone;}
		}

		if(new_state == dsDragDrop) {
			// removeRowCell creates an empty position for checkPosition
			removeRowCell(drag_cell); // user may update drag_cell.Object
			if(can_remove == false) {new_state = dsNone;}
			else if(cell_state == csCenter) {
				Invalidate(); Screen->Cursor = (TCursor) crDrag;
			}
		}
	}
}
//---------------------------------------------------------------------------

void __fastcall TRowGrid::DblClick(void) {
// assuming dblclick only possible with left mouse button
#ifdef _DEBUG
	Exceptions::writeLog("DblClick");
#endif
	if(dblclick && DragState == dsNone) {
		dblclick = false; // see MouseDown
		TCellState dbl_state; RowCell dbl_cell;
		dbl_cell = MouseToCell(drag_x,drag_y,dbl_state);
		if(dbl_state != csNone && dbl_state != csEmpty) {
#ifdef _DEBUG
			AnsiString message("OnDblClickRowCell");
			message += " row:"+IntToStr(dbl_cell.Row);
			message += " start:"+IntToStr(dbl_cell.Start);
			message += " units:"+IntToStr(dbl_cell.Units);
			message += " zorder:"+IntToStr(dbl_cell.Zorder);

			char string[PTRSIZE];
			snprintf(string,PTRSIZE,"%x",dbl_cell.Object);
			message += " object:"+String(string);
			Exceptions::writeLog(message);
#endif
			if(OnDblClickRowCell) {OnDblClickRowCell(this,dbl_cell);}
		}
		// click very often while inserting a zero units cell
		// - mouseup clears DragState, cell_state became csLeft/csRight
		// - dblclick is entered while drag_cell was never inserted.
		// click very often on a cell while holding left-ctrl
		// - rowgrid is refreshed by TFmain::MouseDown
		// - dblclick uses invalid drag_cell object
		// => drag_cell is not valid after mouseup/mousedown/mousemove
		// => mousemove checks dragging before entering state to fix this
		// => mouseup only works with drag_cell if DragState was set
		// => DblClick determines the cell when it is needed to fix this
		// => dblclick var is used for other purposes:
		//    - dblclick after dblclick or dblclick after move
	}
	TControl::DblClick();
}
//---------------------------------------------------------------------------

void __fastcall TRowGrid::MouseDown(TMouseButton Button,TShiftState Shift,
	int X,int Y) {
#ifdef _DEBUG
	Exceptions::writeLog("MouseDown");
#endif
	if((Button == mbLeft || Button == mbRight) && DragState == dsNone &&
		Shift.Contains(ssDouble) == false) { // down, up, dblclick, up, down
		// future: perhaps move MouseToCell to dragRowCell just like DblClick
		dragging = true; drag_hold = false; chk_position = CheckPosition;
		draw_twice = false; drag_button = Button; drag_x = X; drag_y = Y;
		drag_cell = MouseToCell(drag_x,drag_y,cell_state);
		if(AllowAltKey && Shift.Contains(ssAlt)) {
			chk_position = false; cell_state = csEmpty;
		}
		if(cell_state == csEmpty) {
			if(VerticalRows) {Screen->Cursor = (TCursor) crArrowVert;}
			else {Screen->Cursor = (TCursor) crArrowHoriz;}
		}

		dblclick = true;
		// when dblclick, move, down, click, up, dblclick
		// - down ignored since move sets DragState -> dsDragDrop
		// - second dblclick would use drag_cell that was just removed
		// when down, move, click, up, dblclick
		// - DragState: dsNone -> dsDragDrop -> dsNone and dblclick called

		if(AllowShiftKey && Shift.Contains(ssShift)) {
			drag_hold = true; MouseMove(Shift,X,Y);
		}
	}
	TControl::MouseDown(Button,Shift,X,Y);
}
//---------------------------------------------------------------------------

void __fastcall TRowGrid::MouseUp(TMouseButton Button,TShiftState Shift,
	int X,int Y) {
#ifdef _DEBUG
	Exceptions::writeLog("MouseUp");
#endif
	if((Button == mbLeft || Button == mbRight) && drag_hold == false) {
		dragging = false; // after only down, up: move should see a down again
		int old_start,new_start,x;
		if(DragState == dsInsert) {
			MouseToShift(X,Y,old_start,new_start);
			if(new_start < old_start) {
				insertRowCell(drag_cell.Row,new_start,
					old_start-new_start,drag_cell.Zorder,NULL);
				if(can_insert) {ClearDragState();}
			} else if(new_start > old_start) {
				insertRowCell(drag_cell.Row,old_start,
					new_start-old_start,drag_cell.Zorder,NULL);
				if(can_insert) {ClearDragState();}
			} else {ClearDragState();}
			// new_start == old_start: empty cell, do not insert
		} else if(DragState == dsDragDrop || DragState == dsCopyDrop) {
			if(cell_state == csCenter) {
				TCellState drop_state;
				RowCell drop_cell = MouseToCell(X,Y,drop_state);
				if((chk_position == false && drop_state != csNone) ||
					drop_state == csEmpty) {
					// clicked on an empty position
					RowCell old_cell = drag_cell; // for restoring drag_cell
					drag_cell.FRow = drop_cell.Row; // before MouseToShift
					// maybe supply MouseToShift with a row argument instead?
					// try to fit the dragged cell onto the empty position
					if(MouseToShift(X,Y,old_start,new_start)) {
						// restore before possible exception
						drag_cell = old_cell;
						insertRowCell(drop_cell.Row,new_start,
							drag_cell.Units,drag_cell.Zorder,drag_cell.Object);
						if(can_insert) {ClearDragState();}
					} else {drag_cell = old_cell;}
				}
			} else { // border movement, only for dsDragDrop (see dragRowcell)
				// MouseToShift adjusts movement so the dragged cell will fit
				MouseToShift(X,Y,old_start,new_start);
				// keep drag_cell, no need to restore
				RowCell new_cell = drag_cell;
				if(cell_state == csLeft) {
					new_cell.FStart = new_start;
					new_cell.FUnits += old_start-new_start;
				} else {new_cell.FUnits += new_start-old_start;}
				insertRowCell(new_cell.Row,new_cell.Start,
					new_cell.Units,new_cell.Zorder,new_cell.Object);
				if(can_insert) {ClearDragState();}
			}
		}
	} else {drag_hold = false;}
	TControl::MouseUp(Button,Shift,X,Y);
}
//---------------------------------------------------------------------------

RowCell __fastcall TRowGrid::ClearDragState(void) {
#ifdef _DEBUG
	Exceptions::writeLog("ClearDragState");
#endif
	FDragState = dsNone; dragging = false;
	Invalidate(); Screen->Cursor = crDefault;
	return drag_cell;
}
//---------------------------------------------------------------------------
