// $Id: RowGrid.h,v 1.15 2006/09/08 01:15:29 darseq Exp $
// copyright by PostWare (S.B.M. Post)
//---------------------------------------------------------------------------
#ifndef RowGridH
#define RowGridH
//---------------------------------------------------------------------------
#include <SysUtils.hpp>
#include <Classes.hpp>
#include <Controls.hpp>
#include <math.hpp>
#ifdef _USE_OLD_RW_STL
#pragma message "To use the old stl, RowGrid.cpp must be recompiled as well!"
#endif
#include <vector.h>
//---------------------------------------------------------------------------

namespace Rowgrid {
// same namespace as Register

class RowCell {
friend class TRowGrid;
private:
	int FRow;
	int FStart;
	int FUnits;
	int FZorder;
public:
	void * Object;

	__property int Row = {read = FRow};
	__property int Start = {read = FStart};
	__property int Units = {read = FUnits};
	__property int Zorder = {read = FZorder,write = SetZorder};

	bool __fastcall operator ==(const RowCell & RowCell) const {
		return RowCell.Row == Row &&
			RowCell.Start == Start &&
			RowCell.Units == Units &&
			RowCell.Zorder == Zorder &&
			RowCell.Object == Object;
	}

	bool __fastcall operator <(const RowCell & RowCell) const {
		return Zorder < RowCell.Zorder;
	}

	void __fastcall SetZorder(int Value) {if(Value >= 0) {FZorder = Value;}}
};

typedef enum {dsDragDrop,dsCopyDrop,dsInsert,dsNone} TDragState;
typedef enum {csLeft,csCenter,csRight,csEmpty,csNone} TCellState;
typedef void __fastcall (__closure * TInsertRowCellEvent)
	(TObject* Sender,RowCell & RowCell,bool & CanInsert);
typedef void __fastcall (__closure * TRemoveRowCellEvent)
	(TObject* Sender,RowCell & RowCell,bool & CanRemove);
typedef void __fastcall (__closure * TClearRowCellEvent)
	(TObject* Sender,RowCell & RowCell,bool & CanClear);
// following events provide RowCell copies instead of references
typedef void __fastcall (__closure * TDrawRowLineEvent)
	(TObject* Sender,TRect Rect,int Row);
typedef void __fastcall (__closure * TDrawRowCellEvent)
	(TObject* Sender,RowCell RowCell,TRect Rect);
typedef void __fastcall (__closure * TDragRowCellEvent)
	(TObject* Sender,RowCell RowCell,TCellState CellState,
	bool IsCopy,bool & CanDrag);
typedef void __fastcall (__closure * TDblClickRowCellEvent)
	(TObject* Sender,RowCell RowCell);

typedef vector<RowCell> row_t;
typedef vector<row_t> cells_t;
class PACKAGE TRowGrid : public TCustomControl {
private:
	enum {crArrowLeft=1,crArrowRight,crArrowTop,crArrowBottom,
		crArrowHoriz,crArrowVert};
// properties
	int FUnits;
	int FRowCount;
	int FRowHeight;
	int FRowLineWidth;
	int FUnitWidth;
	int FFrameWidth;
	int FCellFrameWidth;
	int FCellBorderWidth;
	bool FDefaultDrawing;
	bool FCheckPosition;
	bool FCrossBorders;
	bool FVerticalRows;
	bool FAllowAltKey;
	bool FAllowShiftKey;
	int FMinVisible;
	TColor FRowLineColor;
	TColor FCellColor;
	TColor FTickColor;
	TDragState FDragState;
	cells_t FCells;

// events
	TInsertRowCellEvent FOnInsertRowCell;
	TRemoveRowCellEvent FOnRemoveRowCell;
	TClearRowCellEvent FOnClearRowCell;
	TDrawRowLineEvent FOnDrawRowLine;
	TDrawRowCellEvent FOnDrawRowCell;
	TDragRowCellEvent FOnDragRowCell;
	TDblClickRowCellEvent FOnDblClickRowCell;

	int __fastcall GetClientWidth(void) {
		return Width - 2*FrameWidth;
	}
	void __fastcall SetClientWidth(int Value) {
		if(Value >= 0) {Width = Value + 2*FrameWidth;}
	}
	int __fastcall GetClientHeight(void) {
		return Height - 2*FrameWidth;
	}
	void __fastcall SetClientHeight(int Value) {
		if(Value >= 0) {Height = Value + 2*FrameWidth;}
	}
	void __fastcall SetUnits(int Value) {
		if(Value >= 1) {FUnits = Value; Invalidate();}
		if(MinVisible > Units) {MinVisible = Units;}
	}
	int __fastcall GetRowCount(void) {return Cells.size();}
	void __fastcall SetRowCount(int Value) {
		if(Value >= 0) {
			for(int r = Value;r < RowCount;r++) {ClearRow(r);}
			// don't care about ClearRow can_clear, resize anyway
			Cells.resize(Value); FRowCount = Value; Invalidate();
		}
	}
	void __fastcall SetRowHeight(int Value) {
		if(Value >= 0) {
			if(RowHeight) {
				int top = FrameWidth + RowLineWidth;
				int o_top = top + drag_cell.Row * (RowLineWidth+RowHeight);
				int n_top = top + drag_cell.Row * (RowLineWidth+Value);
				if(VerticalRows) {
					drag_x = RoundTo(
						n_top + (drag_x-o_top) * (double) Value/RowHeight,0);
				} else {
					drag_y = RoundTo(
						n_top + (drag_y-o_top) * (double) Value/RowHeight,0);
				}
			}
			FRowHeight = Value; Invalidate();
		}
	}
	void __fastcall	SetRowLineWidth(int Value) {
		if(Value >= 0) {
			int top = FrameWidth + RowLineWidth;
			int o_top = top + drag_cell.Row * (RowLineWidth+RowHeight);
			int n_top = FrameWidth + Value + drag_cell.Row * (Value+RowHeight);
			if(VerticalRows) {drag_x = n_top + (drag_x-o_top);}
			else {drag_y = n_top + (drag_y-o_top);}
			FRowLineWidth = Value; Invalidate();
		}
	}
	void __fastcall SetUnitWidth(int Value) {
		if(Value >= 1) {
			if(UnitWidth) {
				if(VerticalRows) {
					drag_y = RoundTo(drag_y * (double) Value/UnitWidth,0);
				} else {drag_x = RoundTo(drag_x * (double) Value/UnitWidth,0);}
			}
			FUnitWidth = Value; Invalidate();
		}
	}
	void __fastcall SetFrameWidth(int Value) {
		if(Value >= 0) {FFrameWidth = Value; Invalidate();}
	}
	void __fastcall SetCellFrameWidth(int Value) {
		if(Value >= 0) {FCellFrameWidth = Value; Invalidate();}
	}
	void __fastcall SetCellBorderWidth(int Value) {
		if(Value >= 0) {FCellBorderWidth = Value;}
	}
	void __fastcall SetDefaultDrawing(bool Value) {
		FDefaultDrawing = Value; Invalidate();
	}
	void __fastcall SetVerticalRows(bool Value) {
		int swp = drag_x; drag_x = drag_y; drag_y = swp;
		FVerticalRows = Value; Invalidate();
	}
	void __fastcall SetMinVisible(int Value) {
		if(Value > Units) {Value = Units;}
		if(Value >= 0) {FMinVisible = Value;}
	}
	void __fastcall SetRowLineColor(TColor Value) {
		FRowLineColor = Value; Invalidate();
	}
	void __fastcall SetCellColor(TColor Value) {
		FCellColor = Value; Invalidate();
	}
	void __fastcall SetTickColor(TColor Value) {
		FTickColor = Value; Invalidate();
	}
	cells_t & __fastcall GetCells(void) {dragging = false; return FCells;}

// internal
	bool dblclick;
	bool can_insert;
	bool can_remove;
	bool internal_call;
//	RowCell prev_cell;

// drawing
	TRect old_rect;
	bool draw_twice;

// dragging
	bool dragging;
	bool drag_hold;
	bool chk_position;
	int drag_x,drag_y;
	RowCell drag_cell;
	TCellState cell_state;
	TMouseButton drag_button;

	void __fastcall Paint(void);
	void __fastcall PaintCell(RowCell & row_cell);
	void __fastcall PaintRow(int row);
	void __fastcall flipRect(TRect & rect);
	void __fastcall drawFocusRect(TRect & rect);
	bool __fastcall MouseToShift(int X,int Y,int & old_start,int & new_start);
	bool __fastcall checkPosition(int X,int & old_start,int & new_start);
	void __fastcall insertRowCell(int row,int start,int units,int zorder,
		void * Object);
	void __fastcall removeRowCell(RowCell & row_cell);
	void __fastcall dragRowCell(TDragState & new_state);
protected:
public:
	__property Canvas; // make Canvas public
	__property TDragState DragState = {read = FDragState, nodefault};
	__property cells_t Cells = {read = GetCells};

	__fastcall TRowGrid(TComponent* Owner);
	void __fastcall ClearRow(int Row);
	void __fastcall InsertRowCell(int Row,int Start,int Units,int Zorder = 0,
		void * Object = NULL);
	void __fastcall RemoveRowCell(RowCell & RowCell);
	RowCell __fastcall MouseToCell(int X,int Y,TCellState & State);
	RowCell __fastcall ClearDragState(void);

	// Provide some flexibility: allow 'external' calls to these
	// functions or allow them to be reimplemented. To implement
	// event handlers, use the OnEvent versions like: OnClick
	DYNAMIC void __fastcall DblClick(void);
	DYNAMIC void __fastcall MouseMove(TShiftState Shift,int X,int Y);
	DYNAMIC void __fastcall MouseUp(TMouseButton Button,TShiftState Shift,
		int X,int Y);
	DYNAMIC void __fastcall MouseDown(TMouseButton Button,TShiftState Shift,
		int X,int Y);
__published:
// new properties
	__property int ClientWidth = {
		read = GetClientWidth, write = SetClientWidth, nodefault
	};
	__property int ClientHeight = {
		read = GetClientHeight, write = SetClientHeight, nodefault
	};
	__property int Units = {
		read = FUnits, write = SetUnits, default = 36
	};
	__property int RowCount = {
		read = GetRowCount, write = SetRowCount, default = 10
	};
	__property int RowHeight = {
		read = FRowHeight, write = SetRowHeight, default = 20
	};
	__property int RowLineWidth = {
		read = FRowLineWidth, write = SetRowLineWidth, default = 10
	};
	__property int UnitWidth = {
		read = FUnitWidth, write = SetUnitWidth, default = 15
	};
	__property int FrameWidth = {
		read = FFrameWidth, write = SetFrameWidth, default = 2
	};
	__property int CellFrameWidth = {
		read = FCellFrameWidth, write = SetCellFrameWidth, default = 1
	};
	__property int CellBorderWidth = {
		read = FCellBorderWidth, write = SetCellBorderWidth, default = 8
	};
	__property bool DefaultDrawing = {
		read = FDefaultDrawing, write = SetDefaultDrawing, default = true
	};
	__property bool CheckPosition = {
		read = FCheckPosition, write = FCheckPosition, default = true
	};
	__property bool CrossBorders = {
		read = FCrossBorders, write = FCrossBorders, default = true
	};
	__property bool VerticalRows = {
		read = FVerticalRows, write = SetVerticalRows, default = false
	};
	__property bool AllowAltKey = {
		read = FAllowAltKey, write = FAllowAltKey, default = false
	};
	__property bool AllowShiftKey = {
		read = FAllowShiftKey, write = FAllowShiftKey, default = true
	};
	__property int MinVisible = {
		read = FMinVisible, write = SetMinVisible, default = 1
	};
	__property TColor RowLineColor = {
		read = FRowLineColor, write = SetRowLineColor, default = clSkyBlue
	};
	__property TColor CellColor = {
		read = FCellColor, write = SetCellColor, default = clBackground
	};
	__property TColor TickColor = {
		read = FTickColor, write = SetTickColor, default = clBlack
	};

// new events
	__property TInsertRowCellEvent OnInsertRowCell = {
		read = FOnInsertRowCell, write = FOnInsertRowCell
	};
	__property TRemoveRowCellEvent OnRemoveRowCell = {
		read = FOnRemoveRowCell, write = FOnRemoveRowCell
	};
	__property TClearRowCellEvent OnClearRowCell = {
		read = FOnClearRowCell, write = FOnClearRowCell
	};
	__property TDrawRowLineEvent OnDrawRowLine = {
		read = FOnDrawRowLine, write = FOnDrawRowLine
	};
	__property TDrawRowCellEvent OnDrawRowCell = {
		read = FOnDrawRowCell, write = FOnDrawRowCell
	};
	__property TDragRowCellEvent OnDragRowCell = {
		read = FOnDragRowCell, write = FOnDragRowCell
	};
	__property TDblClickRowCellEvent OnDblClickRowCell = {
		read = FOnDblClickRowCell, write = FOnDblClickRowCell
	};

// default properties (used TDrawGrid as a reference)
// TWinControl:
	__property Ctl3D;
	__property ParentCtl3D;
	__property TabOrder;
	__property TabStop;

// TControl:
	__property Align;
	__property Anchors;
	__property BiDiMode;
	__property Color;
	__property Cursor;
	__property DragCursor;
	__property DragKind;
	__property DragMode;
	__property Enabled;
	__property Font;
	__property Height;
	__property HelpContext;
	__property HelpKeyword;
	__property HelpType;
	__property Hint;
	__property Left;
	__property Name;
	__property ParentBiDiMode;
	__property ParentColor;
	__property ParentFont;
	__property ParentShowHint;
	__property PopupMenu;
	__property ShowHint;
	__property Top;
	__property Visible;
	__property Width;

// TComponent:
	__property Tag;

// default events (publish all available events)
// TWinControl:
	__property OnDockDrop;
	__property OnDockOver;
	__property OnEnter;
	__property OnExit;
	__property OnGetSiteInfo;
	__property OnKeyDown;
	__property OnKeyPress;
	__property OnKeyUp;
	__property OnUnDock;

// TControl:
	__property OnCanResize;
	__property OnClick;
	__property OnConstrainedResize;
	__property OnContextPopup;
	__property OnDblClick;
	__property OnDragDrop;
	__property OnDragOver;
	__property OnEndDock;
	__property OnEndDrag;
	__property OnMouseDown;
	__property OnMouseMove;
	__property OnMouseUp;
	__property OnMouseWheel;
	__property OnMouseWheelDown;
	__property OnMouseWheelUp;
	__property OnResize;
	__property OnStartDock;
	__property OnStartDrag;
};
//---------------------------------------------------------------------------
} // end namespace Rowgrid
#if !defined(NO_IMPLICIT_NAMESPACE_USE)
using namespace Rowgrid;
#endif

#endif

